/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Definitions of global data structures and functions.
 */


#ifndef DEFS_H
#define DEFS_H


#include "sysdep.h"


/* if these aren't already defined, this is the time to define them... */
#ifndef NULL
#define NULL    ((char *)0L)
#endif

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif


/* somewhere to put unwanted return values */
extern long nowhere;


/* if non-zero, debugging information is sent to stderr */
extern int trace;

#define TRACE(msg)          { if (trace) fprintf(stderr, msg); }
#define TRACE2(msg, arg)    { if (trace) fprintf(stderr, msg, arg); }


/* an x% chance of returning TRUE */
#define PROB(x)         ((random()%100) < x)


/* range used to store probability weightings */
#define MAX_PROB        512L

/* maximum number of patterns that can exist at a time */
#define MAX_PATTERNS    24



/* this is used to 'freeze time' while processing events. It is updated
 * from the volatile clock() function once every cycle though the main
 * control loop, and then stays unchanged while the playback/recording
 * code is active, preventing problems when the time changes unexpectedly
 * part way through a routine.
 */
extern long my_clock;



/* this structure is used to store the individual notes within a pattern */
typedef struct NOTE 
{
    int pitch;                      /* MIDI pitch value */
    int vel;                        /* note velocity */
    long time;                      /* starting time */
    long length;                    /* length of the note */
} NOTE;



/* a string of notes, which can be dynamically resized */
typedef struct PATTERN 
{
    int channel;                    /* MIDI channel to play the pattern on */
    int length;                     /* number of notes in the pattern */
    int allocated_size;             /* how much memory have we allocated? */
    struct PATTERN *parent;         /* the pattern this one is based on */
    struct MUTATOR *mutator;        /* the mutator which produced this */
    long mutator_i;                 /* the integer mutator parameter */
    NOTE note[0];                   /* the notes themselves */
} PATTERN;



/* this structure is used when playing or recording a pattern */
typedef struct PLAYING_PATTERN 
{
    PATTERN *pattern;               /* the pattern we are playing */
    int pos;                        /* current position in the note list */
    long start_time;                /* when we started playing it */
} PLAYING_PATTERN;



/* this structure is used by various parts of the system, to store lists
 * of items with associated 'value' judgements for each one. The functions 
 * in list.c can be used to randomly select an item out of the list using 
 * probability weightings based on the 'coolness' of each item.
 */
typedef struct LIST_MEMBER 
{
    char *desc;                     /* text description of the item */
    long data;                      /* the data itself, can be int or ptr */
    long coolness;                  /* probablility weighting */
} LIST_MEMBER;



/* functions following this prototype are used to mutate patterns */
#ifdef DJGPP
typedef PATTERN *MUTATE_FUNC(PATTERN *p1, PATTERN *p2, int i);
#else
typedef PATTERN *MUTATE_FUNC();
#endif



/* the mutator functions will always be passed a pattern to mutate in the p1
 * parameter. Some mutators also require a second pattern, in order to combine
 * characteristics from different bits of material, so a second pattern will 
 * also be chosen and will be passed in the p2 parameter to the mutator. Some 
 * mutators also require an integer control parameter, for example the 
 * transpose mutator needs to know the interval to transpose by. These 
 * mutators should set the integer_list pointer to an array of LIST_MEMBER 
 * structures, containing all the possible integer parameters in their data 
 * fields. One of these will be selected based on the weightings in the list, 
 * and its data field will be passed as the i parameter to the mutator.
 */
typedef struct MUTATOR
{
    MUTATE_FUNC *func;
    LIST_MEMBER *integer_list;
} MUTATOR;



/* the macro DEF(function_name, function_arguments) is used to define all
 * external functions, as this provides a mechanism for using ANSI-style
 * prototypes with gcc and K&R-style extern definitions with Mark Williams C.
 */


/*********************************************************
 **************** user interface routines ****************
 *********************************************************/

/* if set, the user interface needs redrawing */
extern int ui_update;

/* the probability weighting list being displayed by the user interface */
extern LIST_MEMBER *view_list;

DEF(void draw_list, (LIST_MEMBER *list, int nested));
DEF(void update_screen,());
DEF(void draw_feedback, (long bend));
DEF(void ui_init, ());
DEF(void ui_cleanup, ());
DEF(void do_ui, (int *finish));


/***************************************************
 **************** stuff from midi.c ****************
 ***************************************************/

/* MIDI file logging variables */
extern char *log_name;
extern FILE *log_file;
extern int log_len;

/* if non-zero, the metronome is turned on */
extern int metronome_flag; 

/* tempo in bpm */
extern int metronome_speed; 

/* General MIDI drum channel and hi-hat sound */
extern int metronome_channel;
extern int metronome_note;

/* the active channel for MIDI record/thru */
extern int midi_channel; 

/* current position of the pitch bend wheel, range -128 to 128 */
extern long pitch_bend; 

/* if non-zero, recorded data will be quantized */
extern int quantize_flag; 

/* the fraction of a beat to quantize to (4 = quarter beats) */
extern int quantize_speed; 

DEF(void note_off, (int note));
DEF(void record_note, (int note, int vel));
DEF(void note_on, (int note, int vel));
DEF(void start_recording, ());
DEF(void quantize, (PATTERN *pattern));
DEF(void stop_recording,());
DEF(int is_recording, ());
DEF(void deal_with_event, ());
DEF(void process_input, ());
DEF(void playback, ());
DEF(void do_metronome, ());
DEF(void do_midi, ());
DEF(void midi_init, ());
DEF(void all_notes_off, ());
DEF(void midi_cleanup, ());
DEF(void adjust_times, (long t));
DEF(void change_midi_channel, (int c));
DEF(void midi_puts, (char *s));
DEF(void midi_putl, (long l)); 
DEF(void midi_putw, (short i)); 
DEF(void midi_putc, (char c)); 
DEF(void midi_put_time, ()); 


/******************************************************
 **************** stuff from pattern.c ****************
 ******************************************************/

/* global pool of patterns (the data field points to the pattern structure) */
extern LIST_MEMBER the_patterns[MAX_PATTERNS*2];

/* the larger this value, the more patterns will be played at a time */
extern int pattern_density;

/* the patterns that are currently playing */
extern PLAYING_PATTERN playing_pattern[MAX_PATTERNS];

DEF(PATTERN *resize_pattern, (PATTERN *p, int length));
DEF(void play_pattern, (PATTERN *p));
DEF(void destroy_pattern, (PATTERN *p));
DEF(void pattern_init, ());
DEF(void destroy_all_patterns, ());
DEF(void pattern_cleanup, ());
DEF(void add_pattern, (PATTERN *p, long coolness));
DEF(void play_patterns, ());
DEF(void kill_patterns, ());
DEF(void worsen_with_age, ());


/***************************************************
 **************** stuff from list.c ****************
 ***************************************************/

DEF(void add_to_list, (LIST_MEMBER *list, long item, long coolness));
DEF(void remove_from_list, (LIST_MEMBER *list, long item));
DEF(int get_list_size, (LIST_MEMBER *list));
DEF(long pick_from_list, (LIST_MEMBER *list, long *coolness));
DEF(long pick_bad_item, (LIST_MEMBER *list));
DEF(void adjust_weighting, (LIST_MEMBER *list, long item, long adjustment));


/*****************************************************
 **************** stuff from mutate.c ****************
 *****************************************************/

/* the larger this value, the faster material will be changed */
extern int mutation_rate;

/* global pool of patterns (the data field points to a mutator structure) */
extern LIST_MEMBER the_mutators[];

/* the mutator routines themselves */
extern MUTATOR inverter, reverser, expression, transposer, modal_transposer,
	       scalic_transposer, copy_mode, stretcher, copy_rhythm,
	       cut_block, paste_blocks, displace_notes, range_changer,
	       thinner, thickener;

DEF(void strip_leading_silence, (PATTERN *p));
DEF(void safe_add_pattern, (MUTATOR *mutator, PATTERN *p1, PATTERN *p2, int i, long coolness));
DEF(void mutate, ());
DEF(void parse_weighting, (char *s, LIST_MEMBER *list, int recurse));
DEF(void load_weightings, (char *fname));
DEF(void save_weightings, (char *fname));


/*****************************************************
 **************** stuff from helper.c ****************
 *****************************************************/

#define TEMP_I_SIZE     1024

/* buffer for storing temporary values. Declared globally to conserve stack
 * space and prevent having to keep dynamically allocating memory whenever
 * some scratch space is needed.
 */
extern int temp_i[TEMP_I_SIZE];

extern int temp_i_size;

DEF(void fill_temp_with_pitches, (PATTERN *p));
DEF(void fill_temp_with_unique_pitches, (PATTERN *p));
DEF(void fill_temp_with_mode, (PATTERN *p));
DEF(int closest_note_in_mode, (int i));
DEF(int median_pitch, (PATTERN *p));
DEF(PATTERN *clone_pattern, (PATTERN *p));


/*******************************************************
 **************** stuff from feedback.c ****************
 *******************************************************/

/* a list of all the patterns that have recently been played */
extern PATTERN *feedback_patterns[][MAX_PATTERNS];

DEF(void adjust_pattern, (PATTERN *p, long adjustment, int adjust_mutator));
DEF(void do_feedback, ());
DEF(void feedback_del_pattern, (PATTERN *p));
DEF(void feedback_init, ());


#endif  /* DEFS_H */

