/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Routines for mutating patterns to produce new material.
 */


#include "defs.h"


/* the larger this value, the faster material will be changed */
int mutation_rate = 8;


/* used to keep track of how long since something was last mutated */
long mutation_counter = 0;


/* mutator functions (the data field is a MUTATOR pointer) */
LIST_MEMBER the_mutators[] =
{
    /* <description>        <data>                       <coolness> */

    { "Transpose",          (long)&transposer,           MAX_PROB/4 },
    { "Modal Transpose",    (long)&modal_transposer,     MAX_PROB/4 },
    { "Scalic Transpose",   (long)&scalic_transposer,    MAX_PROB/4 },
    { "Copy Mode",          (long)&copy_mode,            MAX_PROB/4 },
    { "Copy Rhythm",        (long)&copy_rhythm,          MAX_PROB/4 },
    { "Cut Block",          (long)&cut_block,            MAX_PROB/4 },
    { "Paste Blocks",       (long)&paste_blocks,         MAX_PROB/4 },
    { "Time Stretch",       (long)&stretcher,            MAX_PROB/4 },
    { "Alter Range",        (long)&range_changer,        MAX_PROB/4 },
    { "Displace Notes",     (long)&displace_notes,       MAX_PROB/4 },
    { "Add Notes",          (long)&thickener,            MAX_PROB/4 },
    { "Remove Notes",       (long)&thinner,              MAX_PROB/4 },
    { "Invert",             (long)&inverter,             MAX_PROB/4 },
    { "Reverse",            (long)&reverser,             MAX_PROB/4 },
    { "Expression",         (long)&expression,           MAX_PROB/4 },
    { NULL, 0, 0 }
};



/* strip_leading_silence:
 *  Removes any leading silence from a pattern.
 */
void strip_leading_silence(p)
PATTERN *p;
{
    long gap = 0;
    int c;

    if ((p) && (p->length > 0)) {
	while (gap + ((CLK_TCK * 60) / metronome_speed) <= p->note[0].time)
	    gap += (CLK_TCK * 60) / metronome_speed;

	for (c=0; c<p->length; c++)
	    p->note[c].time -= gap;
    }
}



/* safe_add_pattern:
 *  Some mutations could produce patterns containing illegal notes, ie.
 *  too low or two high for MIDI playback. This routine checks a pattern,
 *  only adding it to the pattern list if it is ok.
 */
void safe_add_pattern(mutator, p1, p2, i, coolness)
MUTATOR *mutator;
PATTERN *p1;
PATTERN *p2;
int i;
long coolness;
{
    int x;
    PATTERN *p;

    p = (*mutator->func)(p1, p2, i);

    if (p) {
	for (x=0; x<p->length; x++) {
	    if ((p->note[x].pitch < 16) || 
		(p->note[x].pitch > 112)) {
		TRACE("rejecting invalid pattern\n");
		destroy_pattern(p);
		return; 
	    }

	    if (p->note[x].vel < 1)
		p->note[x].vel = 1;
	    else if (p->note[x].vel > 127)
		p->note[x].vel = 127;
	}

	p->parent = p1;
	p->mutator = mutator;
	p->mutator_i = i;

	strip_leading_silence(p);

	add_pattern(p, coolness); 
    }
}



/* mutate:
 *  Called once per second to randomly mutate patterns in order to obtain
 *  new material.
 */
void mutate()
{
    PATTERN *p1, *p2;
    MUTATOR *mutator;
    int i;
    long coolness;

    mutation_counter += mutation_rate;

    /* while there are some mutations needing to be done */
    while (mutation_counter >= 8) {

	/* pick a mutation type */
	mutator = (MUTATOR *)pick_from_list(the_mutators, &nowhere);

	if (mutator) {
	    /* pick a pattern to mutate */
	    p1 = (PATTERN *)pick_from_list(the_patterns, &coolness);

	    /* pick a second pattern in case the mutator wants two */
	    p2 = (PATTERN *)pick_from_list(the_patterns, &nowhere);

	    if ((p1) && (p2)) {
		if (mutator->integer_list)
		    /* pick an integer parameter if one is required */
		    i = pick_from_list(mutator->integer_list, &nowhere);
		else
		    i = 0;

		/* and finally do the mutation */
		safe_add_pattern(mutator, p1, p2, i, coolness);
	    } 
	} 

	mutation_counter -= 8;
    }
}



/* parse_weighting:
 *  Tries to find an entry in the list corresponding to the start of s,
 *  and adjust its weighting accordingly. If recurse is set, strings in
 *  the name.name format will be interpreted as multi-level mutator lists.
 */
void parse_weighting(s, list, recurse)
char *s;
LIST_MEMBER *list;
int recurse;
{
    char *p;
    MUTATOR *m;

    while (list->data) {
	if (strncmp(s, list->desc, strlen(list->desc)) == 0) {
	    p = s + strlen(list->desc);
	    if ((p[0] == ':') && (p[1] == ' ')) {
		list->coolness = atol(p+2);
		return;
	    }
	    else if ((recurse) && (p[0] == '.')) {
		m = (MUTATOR *)list->data;
		if (m->integer_list)
		    parse_weighting(p+1, m->integer_list, FALSE);
		return;
	    }
	}

	list++;
    }
}



/* load_weightings:
 *  Reads probability weightings from a file.
 */
void load_weightings(fname)
char *fname;
{
    char buf[80];
    FILE *f = fopen(fname, "r");

    if (f) {
	while (fgets(buf, 79, f))
	    parse_weighting(buf, the_mutators, TRUE);

	fclose(f); 
    }
}



/* save_weightings:
 *  Writes probability weightings into a file.
 */
void save_weightings(fname)
char *fname;
{
    int c, c2;
    FILE *f = fopen(fname, "w");
    LIST_MEMBER *l;

    if (f) {
	for (c=0; the_mutators[c].data; c++) {
	    fprintf(f, "%s: %ld\n", the_mutators[c].desc, 
				    the_mutators[c].coolness);

	    l = ((MUTATOR *)the_mutators[c].data)->integer_list;
	    if (l) {
		for (c2=0; l[c2].data; c2++)
		    fprintf(f, "%s.%s: %ld\n", the_mutators[c].desc, 
						l[c2].desc, l[c2].coolness);
	    }

	    fprintf(f, "\n");
	}

	fclose(f); 
    }
}
