/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  DJGPP-specific MS-DOS routines, using the Allegro library to access
 *  the MPU-401 or SB MIDI interfaces.
 */


#ifndef DJGPP
#error This code is machine-dependent. It will only work with djgpp...
#endif


#include <dos.h>
#include <sys/exceptn.h>
#include <allegro.h>

#include "defs.h"



#define MIDI_BUFFER_SIZE    8192


/* ring buffer for storing MIDI input */
volatile int midi_buffer_head, midi_buffer_tail;
volatile unsigned char midi_buffer[MIDI_BUFFER_SIZE];



/* midi_handler:
 *  Interrupt handler for the MIDI data received message.
 */
void midi_handler(unsigned char data)
{
    midi_buffer[midi_buffer_tail] = data;
    midi_buffer_tail++;
    if (midi_buffer_tail >= MIDI_BUFFER_SIZE)
	midi_buffer_tail = 0;
}

END_OF_FUNCTION(midi_handler);



/* midi_waiting:
 *  Returns non-zero if MIDI data is waiting to be read.
 */
int midi_waiting()
{
    if (midi_buffer_head == midi_buffer_tail)
	return 0;
    else
	return 1;
}



/* midi_in:
 *  Reads a byte from the MIDI interface, or returns zero if no data
 *  is available.
 */
unsigned char midi_in()
{
    unsigned char ret;

    if (midi_buffer_head == midi_buffer_tail)
	return 0;

    ret = midi_buffer[midi_buffer_head];
    midi_buffer_head++;
    if (midi_buffer_head >= MIDI_BUFFER_SIZE)
	midi_buffer_head = 0;

    return ret; 
}



/* midi_output:
 *  Writes a byte to the MIDI interface. To make this work with Allegro
 *  drivers other than just the raw MIDI ones, we need to cache a number
 *  of bytes and only transmit them when a complete MIDI command sequence
 *  is available.
 */
void midi_output(unsigned char byte)
{
    static unsigned char buf[16];
    static int count = 0;

    if (byte & 0x80) {
	if (count)
	    midi_out(buf, count);

	buf[0] = byte;
	count = 1;
    }
    else {
	buf[count++] = byte;
	if (count >= 3) {
	    midi_out(buf, count);
	    count = 0;
	}
    }
}



/* hw_init:
 *  Sets up all the MS-DOS hardware handlers.
 */
int hw_init()
{
    __djgpp_set_ctrl_c(0);
    setcbrk(0);

    allegro_init();

    if (install_sound(DIGI_NONE, MIDI_AUTODETECT, NULL) != 0) {
	printf("\nError initialising sound\n%s\n\n", allegro_error);
	return -1;
    }

    if (install_sound_input(DIGI_NONE, MIDI_AUTODETECT) != 0) {
	printf("\nError initialising sound input\n%s\n\n", allegro_error);
	return -1;
    }

    LOCK_VARIABLE(midi_buffer_head);
    LOCK_VARIABLE(midi_buffer_tail);
    LOCK_VARIABLE(midi_buffer);
    LOCK_FUNCTION(midi_handler);

    midi_recorder = midi_handler;

    return 0;
}



/* hw_cleanup:
 *  Cleans up at the end of the program.
 */
void hw_cleanup()
{
    allegro_exit();
}
