/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Mutation routines for transposing patterns and copying scales from one
 *  pattern to another.
 */


#include "defs.h"


MUTATE_FUNC trans_func, modal_trans_func, 
	    scalic_trans_func, copy_mode_func;


/* initially biased in favour of octave transposition */
LIST_MEMBER trans_params[] =
{
    /* <description>                <data>                       <coolness> */
    { "1 semitone",                 1,                           MAX_PROB/16 },
    { "2 semitones",                2,                           MAX_PROB/16 },
    { "3 semitones",                3,                           MAX_PROB/16 },
    { "4 semitones",                4,                           MAX_PROB/16 },
    { "5 semitones",                5,                           MAX_PROB/16 },
    { "6 semitones",                6,                           MAX_PROB/16 },
    { "7 semitones",                7,                           MAX_PROB/16 },
    { "8 semitones",                8,                           MAX_PROB/16 },
    { "9 semitones",                9,                           MAX_PROB/16 },
    { "10 semitones",               10,                          MAX_PROB/16 },
    { "11 semitones",               11,                          MAX_PROB/16 },
    { "12 semitones",               12,                          MAX_PROB/2 },
    { NULL, 0, 0 }
};



MUTATOR transposer =
{
    trans_func,                 /* literally transposes a pattern */
    trans_params                /* by how much? */
};



MUTATOR modal_transposer =
{
    modal_trans_func,           /* transposes a pattern within a mode */
    trans_params                /* by how much? */
};



/* initially biased in favour of single step shifts */
LIST_MEMBER scalic_trans_params[] =
{
    /* <description>                <data>                       <coolness> */
    { "-3 steps",                   -3,                          MAX_PROB/16 },
    { "-2 steps",                   -2,                          MAX_PROB/16 },
    { "-1 step",                    -1,                          MAX_PROB/2 },
    { "1 step",                     1,                           MAX_PROB/2 },
    { "2 steps",                    2,                           MAX_PROB/16 },
    { "3 steps",                    3,                           MAX_PROB/16 },
    { NULL, 0, 0 }
};



MUTATOR scalic_transposer =
{
    scalic_trans_func,          /* transposes a pattern within a scale */
    scalic_trans_params         /* by how much? */
};



MUTATOR copy_mode=
{
    copy_mode_func,             /* transposes a pattern within a mode */
    NULL                        /* don't need an integer parameter */
};



/* trans_func:
 *  Transposes a pattern.
 */
PATTERN *trans_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;

    if (PROB(50))
	i = -i;

    TRACE2("trans_func(%d)\n", i);

    p = clone_pattern(p1);

    if (p)
	for (c=0; c<p->length; c++)
	    p->note[c].pitch += i;

    return p;
}



/* modal_trans_func:
 *  Transposes a pattern without changing its mode.
 */
PATTERN *modal_trans_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;

    if (PROB(50))
	i = -i;

    TRACE2("modal_trans_func(%d)\n", i);

    p = clone_pattern(p1);

    if (p) {
	fill_temp_with_mode(p);

	for (c=0; c<p->length; c++)
	    p->note[c].pitch = closest_note_in_mode(p->note[c].pitch + i);
    }

    return p;
}



/* scalic_trans_func:
 *  Transposes a pattern without changing its scale.
 */
PATTERN *scalic_trans_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int x, c, l;
    int dir;
    int shift;
    int range;

    TRACE2("scalic_trans_func(%d)\n", i);

    p = clone_pattern(p1);

    if (p) {
	fill_temp_with_unique_pitches(p);

	/* calculate range of the scale, rounded up to next octave */
	range = temp_i[temp_i_size-1] - temp_i[0];
	while (range % 12)
	    range++;

	if (i < 0) {
	    i = -i;
	    dir = -1;
	}
	else
	    dir = 1;

	for (x=0; x<p->length; x++) {
	    /* find the position in the scale */
	    for (c=0; c<temp_i_size-1; c++)
		if (temp_i[c] == p->note[x].pitch)
		    break;

	    shift = 0;

	    for (l=0; l<i; l++) {
		c += dir;
		if (c < 0) {
		    shift -= range;
		    c = temp_i_size - 1;
		    /* special case if lowest and highest notes are in 8ves */
		    if ((c > 0) && ((temp_i[c]%12) == (temp_i[0]%12)))
			c--;
		}
		else if (c >= temp_i_size) {
		    shift += range;
		    /* special case if lowest and highest notes are in 8ves */
		    if ((temp_i_size>1) && 
			((temp_i[0]%12) == (temp_i[temp_i_size-1]%12)))
			c = 1;
		    else
			c = 0;
		}
	    }

	    p->note[x].pitch = temp_i[c] + shift;
	}
    }

    return p;
}



/* copy_mode_func:
 *  Forces a pattern into the mode used by a different pattern.
 */
PATTERN *copy_mode_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;

    TRACE("copy_mode_func()\n");

    p = clone_pattern(p1);

    if (p) {
	fill_temp_with_mode(p2);

	for (c=0; c<p->length; c++)
	    p->note[c].pitch = closest_note_in_mode(p->note[c].pitch);
    }

    return p;
}


